/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */

#if MPI
#ifndef _DECOMP_H_
#define _DECOMP_H_
#ifdef _OWN
#undef _OWN
#define OWN
#endif
#include "extern.h"

const static int DECOMP_VOLUME_ADJUST = 1;
const static int DECOMP_SURFACE_AREA_ADJUST = 100;

/* 8< decomposition_datatype */
typedef struct{
	int nx, ny, nz;
} Decomposition;
/* >8 decomposition_datatype */

/*! Computes the dimensions of the worst (i.e. largest) subdomain of a given decomposition.
 *	\param d Decomposition in which to find the subdomain.
 *	\param xlen Address at which the x dimension of the subdomain should be stored.
 *	\param ylen Address at which the y dimension of the subdomain should be stored.
 *	\param zlen Address at which the z dimension of the subdomain should be stored.
 */
void decomp_getWorstSubdomainDimensions(Decomposition d, int *xlen, int *ylen, int *zlen);

/*! Computes the largest volume of the any subdomain in a given decomposition.
 *	\param d Decomposition in which to find the subdomain.
 *	\return Volume of the subdomain (pt^3).
 */
int decomp_getWorstSubdomainVolume(Decomposition d);

/*! Computes the largest surface area of the any subdomain in a given decomposition.
 *	\param d Decomposition in which to find the subdomain.
 *	\return Surface area of the subdomain (pt^2).
 */
int decomp_getWorstSubdomainSurfaceArea(Decomposition d);

/*! Computes metric for comparison of decompositions.
 *	\param d Decomposition to be evaluated.
 *	\return Metric value. Smaller is better.
 */
float decomp_getMetric(Decomposition d);

/*! Compares two decompositions using decomp_getMetric().
 *	\param a First decomposition.
 *	\param b Second decomposition.
 *	\return The better of a and b, according to decomp_getMetric().
 *	\sa decomp_getMetric
 */
Decomposition decomp_getBest(Decomposition a, Decomposition b);

/*! Computes a preferred domain decomposition for a given number of processes.
 *	\param num_procs The number of available processes.
 *	\param nx Address at which the x dimension of the supergrid should be stored.
 *	\param ny Address at which the y dimension of the supergrid should be stored.
 *	\param nz Address at which the z dimension of the supergrid should be stored.
 *	\return The number of active processes in the preferred decomposition.
 */
int decompose(int num_procs, int *nx, int *ny, int *nz);

/*! Computes the superindices of a subdomain, given its rank. 
 *	\param rank Rank of the process for which superindices are to be calculated.
 *	\param ix Address at which the x supergrid coordinate should be stored.
 *	\param iy Address at which the y supergrid coordinate should be stored.
 *	\param iz Address at which the z supergrid coordinate should be stored.
*/
void decomp_getSuperindices(int rank, int *ix, int *iy, int *iz);

/*! Computes the local subdomain's dimensions. 
 *	Results for all subdomains are stored in subdomain_limits.
 *	Results for local subdomain are stored in local_xmin, local_xmax, etc.
*/
int decomp_getSubdomainDimensions();

/*! Computes the local subdomain's size (including boundary/halo points). 
 *	Results are stored in xlen, ylen and zlen.
*/
void decomp_getSubdomainSize();

/*! Defines datatypes for halo swapping.
*/
void decomp_defineHaloTypes();

/*! Defines a global communicator over all active processes. 
 *	\param old_rank The process's rank in MPI_COMM_WORLD.
 *	\param new_comm Address at which the new communicator will be created.
 *	\param new_rank Address at which the process's rank in new_comm will be stored. Will be equal to old_rank if the process is idle.
 *	\return 1 if the process is idle in the decomposition.
*/
int decomp_globalCommunicator(int old_rank, MPI_Comm *new_comm, int *new_rank);

#endif /* end of include guard: _DECOMP_H_ */

#endif /* MPI */
