/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <math.h>
#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include "beatbox.h"
#include "error.h"
#include "state.h"
#define OWN
#include "geometry.h"
#undef OWN
#include "mpi_io_choice.h"

/*  Shifts required to put first tissue point at ONE,TWO,TRI. */
static int x_offset,y_offset,z_offset;

/** 
 * Validates vector components. 
 * Returns 1 if the components describe a unit vector, 0 otherwise.
 */
static int isUnitVector(double v1, double v2, double v3, double tolerance){
  double error = fabs(sqrt((v1*v1) + (v2*v2) + (v3*v3)) -1 );
  if (error > tolerance || isnan(error)){
    return 0;
  }
  return 1;
}

/** 
 * Normalises vector components to produce a unit vector.
 */
static int normaliseVector(double *v1, double *v2, double *v3){
  double length = sqrt((*v1 * *v1 + *v2 * *v2 + *v3 * *v3));
  if(length <= UNIT_VECTOR_TOLERANCE) return 0;
  *v1 = *v1 / length;
  *v2 = *v2 / length;
  *v3 = *v3 / length;
  return 1;
}

/* 8< populatemesh */
/**
 * Loads geometry data from file into Geom.
 */
int populateMesh(FILE *geomFile, const char *geomFileName, int normaliseVectors) {
  int line = 0;
  int status;
  int x_in,y_in,z_in; /* Coords according to the geometry file. */
  int x,y,z; /* Corrected coords, based on offsets. */
  int x0,x1,y0,y1,z0,z1;
  double fibre1, fibre2, fibre3;
  int i;
  int checkVector = 0;
  int invalidVectors = 0;
  int tissueInHalos = 0;
  
  /* Set bounds for seq/MPI versions */
#if MPI
  /*  Include halos, since we'll need to know if they're tissue too. */
  x0 = local_xmin - 1;
  x1 = local_xmax + 1;
  y0 = local_ymin - 1;
  y1 = local_ymax + 1;
  z0 = local_zmin - 1;
  z1 = local_zmax + 1;
#else
  x0 = 0;
  x1 = xmax;
  y0 = 0;
  y1 = ymax;
  z0 = 0;
  z1 = zmax;
#endif
  
  /*  Zero all points, since files are sparse. */
  for (x=x0; x<x1; x++) {
    for (y=y0; y<y1; y++) {
      for (z=z0; z<z1; z++) {
	Geom[ geom_ind(x, y, z, GEOM_STATUS  ) ] = GEOM_VOID;
	Geom[ geom_ind(x, y, z, GEOM_FIBRE_1 ) ] = 0.0;
	Geom[ geom_ind(x, y, z, GEOM_FIBRE_2 ) ] = 0.0;
	Geom[ geom_ind(x, y, z, GEOM_FIBRE_3 ) ] = 0.0;
      }
    }
  }
  
  MESSAGE("\tLoading geometry data"); 
  if (normaliseVectors) {MESSAGE(" and normalising vectors");}
  MESSAGE("...");
  if(fflush_or_not==1) fflush(stdout);
  rewind(geomFile);
  while (fscanf(geomFile, 
		"%d,%d,%d,%d,%lf,%lf,%lf\n",
		&x_in, &y_in, &z_in, &status, &fibre1, &fibre2, &fibre3)
	 != EOF) {
    
    /* Validate status */
    if (!(status==GEOM_VOID || status==GEOM_TISSUE || status==GEOM_TISSUE2)) 
      EXPECTED_ERROR3("Problem reading geometry file. Invalid status value %d given on line %d of %s\n.", 
		      status, line, geomFileName);
    
    /* Shift coords */
    x = x_in + x_offset;
    y = y_in + y_offset;
    z = z_in + z_offset;
    
    /* Ensure all adjusted coords are inside containing box. */
    assert(x>=0 && x<xmax);
    assert(y>=0 && y<ymax);
    assert(z>=0 && z<zmax);
    
    assert(!(isnan(fibre1)));
    assert(!(isnan(fibre2)));
    assert(!(isnan(fibre3)));
    
    /* Don't need to do anything if the point is void */
    if (status == GEOM_TISSUE || status == GEOM_TISSUE2) {
      
      /* Ensure no tissue points in global halos */
      if (dim>=1) assert(x!=0 && x!=xmax-1);
      if (dim>=2) assert(y!=0 && y!=ymax-1);
      if (dim==3) assert(z!=0 && z!=zmax-1);

      /* Check and optionally normalise fibre vectors if anisotropy is on. */
      if (normaliseVectors && ANISOTROPY_ON && !normaliseVector(&fibre1,&fibre2,&fibre3)) 
	EXPECTED_ERROR6("Error normalising vector at (%d,%d,%d): [%lf,%lf,%lf]. Vector is too short.",
			x_in,y_in,z_in,fibre1,fibre2,fibre3);
      
      /* Check vectors, even if we're normalising.
	 2D meshes may cause errors here, if the only non-zero vector component is along the z axis. */
      if (ANISOTROPY_ON && !isUnitVector(fibre1,fibre2,fibre3,UNIT_VECTOR_TOLERANCE)) {
	MESSAGE("Fibre vector at point (%d,%d,%d): [%lf,%lf,%lf] is not a valid unit vector.\n",
		x_in,y_in,z_in,fibre1,fibre2,fibre3);
	invalidVectors = 1;
      }
      
      /* If it belongs to the local subdomain assign the point */
      if (x>=x0 && x<x1 &&
	  y>=y0 && y<y1 &&
	  z>=z0 && z<z1) {
	Geom[ geom_ind(x, y, z, GEOM_STATUS ) ] = status;
	Geom[ geom_ind(x, y, z, GEOM_FIBRE_1) ] = fibre1;
	Geom[ geom_ind(x, y, z, GEOM_FIBRE_2) ] = fibre2;
	Geom[ geom_ind(x, y, z, GEOM_FIBRE_3) ] = fibre3;
      }
    } /* status != GEOM_VOID */
    
    line++;
  } /* while (fscanf(geomFile)) */
  
  if(invalidVectors) 
    EXPECTED_ERROR0("Invalid fibre vectors encountered. "
		    "Try using 'normaliseVectors=1' in state to correct them.\n");

  MESSAGE("done.\n",line);
  if(fflush_or_not==1) fflush(stdout);
  return 1;
}
/* >8 populatemesh */

/* 8< getmeshdimensions */
int getMeshDimensions (FILE *geomFile, int *mesh_xmax,int *mesh_ymax,int *mesh_zmax) {
  int first,
    x,y,z,status,
    min_x,max_x,min_y,max_y,min_z,max_z,
    geom_xlen, geom_ylen, geom_zlen;
  double fibre1, fibre2, fibre3;
  
  first = 1;
  rewind(geomFile);
  while (fscanf(geomFile,
		"%d,%d,%d,%d,%lf,%lf,%lf\n",
		&x, &y, &z, &status, &fibre1, &fibre2, &fibre3)
	 != EOF) {
    if (first) {
      min_x = max_x = x;
      min_y = max_y = y;
      min_z = max_z = z;
      first = 0;
    } else if (status == GEOM_TISSUE || status == GEOM_TISSUE2) {
      min_x = x<min_x?x:min_x;
      max_x = x>max_x?x:max_x;
      min_y = y<min_y?y:min_y;
      max_y = y>max_y?y:max_y;
      min_z = z<min_z?z:min_z;
      max_z = z>max_z?z:max_z;
    }
  }
  
  geom_xlen = ((max_x - min_x) + 1);
  geom_ylen = ((max_y - min_y) + 1);
  geom_zlen = ((max_z - min_z) + 1);
  
  /* How should we shift the incoming coords to 
   * put the first tissue point at 1,1,1?	*/
  x_offset = ((geom_xlen>1?1:0)-min_x);
  y_offset = ((geom_ylen>1?1:0)-min_y);
  z_offset = ((geom_zlen>1?1:0)-min_z);
  
  *mesh_xmax = geom_xlen + (geom_xlen>1?2:0); /* Num points + bounds. */
  *mesh_ymax = geom_ylen + (geom_ylen>1?2:0); /* Num points + bounds. */
  *mesh_zmax = geom_zlen + (geom_zlen>1?2:0); /* Num points + bounds. */
  
  if(fflush_or_not==1) fflush(stdout);
  return 1;
}
/* >8 getmeshdimensions */

