/**
 *  Copyright (C) (2010-2013) Vadim Biktashev, Alexander Karpov, 
 *  Irina Biktasheva, Ross McFarlane, Sanjay Kharche,
 *  Mario Antonioletti, Adrian Jackson.
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


#if MPI
#include <mpi.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "system.h"
#include "beatbox.h"
#include "k_.h"
#include "device.h"
#include "qpp.h"
#include "bikt.h"
#include "decomp.h"
#include "error.h"
#include "geometry.h"
#define OWN
#include "state.h"
#undef OWN
#include "mpi_io_choice.h"

static real  *state_u1;
static real  *geom;
static int already_created=0;
typedef struct {
  /* Geometry file */
  char geometryname[MAXPATH];
  FILE *geometry;
  /* Geometry options */
  int normaliseVectors;
  int correctBoundaries;
  int anisotropy;
  
  /*  Global bounds */
  int xmax, ymax, zmax, vmax;
} STR;

/* state is performed once just when reading from parameter file */
int state (char *w) {
  /*  Loop iterator */
  int i;
  /* Index from which shifts are measured. */
  int base;
  
  STR *S= (STR *) Malloc(sizeof(STR));
  
  if (!S) ERROR0("not enough memory");
	
  if (already_created) ERROR0("state must be defined only once");
  
  ACCEPTF(geometry,"r","");
  GEOMETRY_ON = (S->geometry!=NULL);
  ACCEPTI(anisotropy,0,0,1);
  ANISOTROPY_ON = S->anisotropy && GEOMETRY_ON; /* Anisotropy is a geometry-only feature */
	
  if (GEOMETRY_ON) {

    ACCEPTI(normaliseVectors,0,0,1);
    /*  Dimensions of mesh's containing box. */
    int mesh_xmax,mesh_ymax,mesh_zmax;
    
    MESSAGE("\tUsing geometry from %s.\n",S->geometryname);
    if (stateDimensionsExist(w)) {
      MESSAGE("\n/**************************************************** WARNING ****************************************************\n");
      MESSAGE("Dimensions of the simulation medium (xmax,ymax,zmax) cannot be specified when geometry is in use.\n");
      MESSAGE("The dimensions provided will be ignored.\n");
      MESSAGE("*****************************************************************************************************************/\n\n");
    }

    if(ANISOTROPY_ON) MESSAGE("\tAnisotropy is on.\n");

    getMeshDimensions(S->geometry,&mesh_xmax,&mesh_ymax,&mesh_zmax);
    xmax = S->xmax = mesh_xmax;
    ymax = S->ymax = mesh_ymax;
    zmax = S->zmax = mesh_zmax;
    geom_vmax = 4; /* STATUS + 3 VECTOR COMPONENTS */
    
  } else {

    ACCEPTI(xmax,INONE,1,INONE); xmax=S->xmax;
    ACCEPTI(ymax,INONE,1,INONE); ymax=S->ymax;
    ACCEPTI(zmax,    1,1,INONE); zmax=S->zmax;
    if (S->anisotropy) {
      MESSAGE("\n/**************************************************** WARNING ****************************************************\n");
      MESSAGE("Anisotropy can only be activated when geometry is in use.\n");
      MESSAGE("We'll continue without for now.\n");
      MESSAGE("*****************************************************************************************************************/\n\n");
    }

  }

  ACCEPTI(vmax,    2,2,INONE); vmax=S->vmax;
  MESSAGE("\tSimulation medium (%d x %d x %d)\n",xmax,ymax,zmax);
  if (fflush_or_not==1)
    fflush(stdout);
	
  /* If any dimension is exactly 2, complain.
   * Since it's not a single point, we'll need diffusion,
   * but we don't have enough points for boundaries.
   */
  if (xmax==2||ymax==2||zmax==2) {
    ERROR0("One or more dimensions (xmax,ymax,zmax) is equal to 2. Dimensions must be either 1 or >=3.");
  }
  
  /* If the dimension is in use, flag it so we can 
     treat the first and last points as boundaries.*/	
  TRI = zmax >=3 ? 1 : 0;
  TWO = ymax >=3 ? 1 : 0;
  ONE = xmax >=3 ? 1 : 0;
  dim = ONE + TWO + TRI;
  
  if(dim==1 && !ONE)		ERROR0("1-Dimensional simulations must be defined on the x axis.");
  if(dim==1 && GEOMETRY_ON)	ERROR0("Geometry requires a 2- or 3-dimensional simulation medium.");
  if(dim==2 && !(ONE && TWO))	ERROR0("2-Dimensional simulations must be defined on the x and y axes.");
	
#if MPI /*  MPI ONLY */
  /*  Get total number of processes. */
  int mpi_size;
  mpi_errno = MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
  CHECK_MPI_SUCCESS("Couldn't get number of MPI processes.")
	
  /* Compute domain decomposition. */
  num_active_procs = decompose(mpi_size, &mpi_nx, &mpi_ny, &mpi_nz);
  MESSAGE("\tDomain decomposed as (%03d,%03d,%03d).\n",mpi_nx,mpi_ny,mpi_nz);
  MESSAGE("\t%d processors of %d remain idle.\n",mpi_size-num_active_procs, mpi_size);

  /*  Define communicator over active processes. If we're idle, we're done. */
  I_AM_IDLE = decomp_globalCommunicator (mpi_rank, &ALL_ACTIVE_PROCS, &mpi_rank);

  if (I_AM_IDLE) {
    printf("IAMIDLE: %d %d %d\n",I_AM_IDLE,mpi_rank,mpi_size);
    I_AM_IDLE = 1; // why: I_AM_IDLE is already non zero.
    FREE(S);
    return (1);
  }
  /**************************************************************************
   *************************************************************************/
	
  /*  Compute superindices */
  decomp_getSuperindices(mpi_rank, &mpi_ix, &mpi_iy, &mpi_iz);
  decomp_getSubdomainDimensions();
  decomp_getSubdomainSize();
  decomp_defineHaloTypes();

  // what is xlen, ylen, zlen by the time we get here? It should be xmax/num_size or close to that.
		
#else /*  SEQUENTIAL ONLY */
  /*  Size of the medium, including boundaries. */
  xlen = xmax;
  ylen = ymax;
  zlen = zmax;	
#endif

  printf("Components of the state: %d %d %d %d \n",xlen,ylen,zlen,vmax);
  /* This is the actual declaration of memory */
  int num_values = xlen*ylen*zlen*vmax; // what value is in num_values? It should be total points/mpi_size
  /* numvalues on thread 1 is 306 for fhn1.bbs: so the CALLOC is declaring a 306 element 1D array.
     It is undefined on all other threads */
  CALLOC(state_u1,num_values,sizeof(real));
	
  /*  Zero medium to ensure consistency */
  for(i=0; i<num_values; i++){
    state_u1[i] = 0.0;
  }
	
  /*  Get constants for index computation. */
  vmax_zmax = vmax * zlen;
  vmax_zmax_ymax = vmax * zlen * ylen;
  vmax_ymax = vmax * ylen;
	
  /*  Get constants for geometry index computation. */
  if (GEOMETRY_ON) {
    geom_vmax_zmax= geom_vmax * zlen;
    geom_vmax_zmax_ymax= geom_vmax * zlen * ylen;
    geom_vmax_ymax = geom_vmax * ylen;
  }

/*  Shift increments */
#if MPI
  /*  Computed from local points to avoid upsetting ind() */
  base = ind(local_xmin,	local_ymin,		local_zmin,		0);
  DX   = ind(local_xmin+1,local_ymin,		local_zmin,		0) - base;
  DY   = ind(local_xmin,	local_ymin+1,	local_zmin,		0) - base;
  DZ   = ind(local_xmin,	local_ymin,		local_zmin+1,	0) - base;
  DV   = ind(local_xmin,	local_ymin,		local_zmin,		1) - base;
#else /*  Sequential */
  base = ind(0,0,0,0);
  DX   = ind(1,0,0,0) - base;
  DY   = ind(0,1,0,0) - base;
  DZ   = ind(0,0,1,0) - base;
  DV   = ind(0,0,0,1) - base;
#endif

  New = state_u1;
	
  if (GEOMETRY_ON) {
    int num_geom_values = xlen*ylen*zlen*geom_vmax;
    CALLOC(geom,num_geom_values,sizeof(real));
    Geom = geom;
    if (!populateMesh(S->geometry,S->geometryname,S->normaliseVectors)) return 0;
    fclose(S->geometry);
  }
	
  FREE(S);
  return (1);
	
} /*  state() */

void state_free(void) {
  FREE(New);
  if(GEOMETRY_ON) FREE(Geom);
}


/**************************************************************************
 ************************** MPI FUNCTIONS *********************************
 **************************************************************************/
#if MPI

int getProcessRank (int ix,int iy,int iz) {
  return (ix*mpi_ny*mpi_nz)+(iy*mpi_nz)+iz;
}


int getRankContainingPoint(int x,int y,int z) {
  int rank,ix,iy,iz;
  int inx,iny,inz;
  for (rank=0;rank<num_active_procs;rank++) {
    decomp_getSuperindices (rank,&ix,&iy,&iz);
    inx = ( x >= subdomain_limits[rank].local_xmin || ix == 0 ) && ( x <  subdomain_limits[rank].local_xmax	|| ix == mpi_nx-1 );
    iny = ( y >= subdomain_limits[rank].local_ymin || iy == 0 ) && ( y <  subdomain_limits[rank].local_ymax	|| iy == mpi_ny-1 );
    inz = ( z >= subdomain_limits[rank].local_zmin || iz == 0 ) && ( z <  subdomain_limits[rank].local_zmax	|| iz == mpi_nz-1 );
    if (inx && iny && inz) {
      return rank;
    }
  }
  return -1;
}
#endif
