/**
 *  Copyright (C) (2010-2013) Vadim Biktashev, Alexander Karpov, 
 *  Irina Biktasheva, Ross McFarlane, Sanjay Kharche,
 *  Mario Antonioletti, Adrian Jackson.
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef _STATE_H_
#define _STATE_H_ 

#ifdef OWN
#undef OWN
#define _OWN
#endif
#include "extern.h"
#include "k_.h"
#ifdef MPI
#include <mpi.h>
#endif /*	MPI	*/
#ifdef _OWN
#undef _OWN
#define OWN
#endif
#include "extern.h"
/* as some state vars are to be k-vars */

/*
 * All items to be saved for continuing calculations later are here,
 * except k-vars
 */

/* Pointer to current state Ordered V,Z,Y,X, so that steps between
 * neighbouring cells in the x-axis are largest.  This means that a
 * y/z plane is represented by a contiguous block of memory.  This
 * declaration means that New is a global EXTERN variable. Need to
 * "gatherv this variable on master node.  my_rank = 0 to be treated
 * as master node.
 */

EXTERN real *New;		

/**
 *	Pointer to geometry data.
 *	Separated from New to prevent confusions over vmax.
 */
EXTERN real *Geom;

EXTERN INT t;				/* device loops counter */
EXTERN INT vmax,xmax,ymax,zmax;		/* the sizes of a state */
EXTERN int xlen,ylen,zlen;		/* Lengths of axes in memory (i.e. including boundaries) */
EXTERN INT GEOMETRY_ON;			/* is geometry on?		*/
EXTERN INT geom_vmax;			/* number of geom variables */

EXTERN INT ANISOTROPY_ON;		/* is anisotropy on?		*/

EXTERN long vmax_zmax, vmax_zmax_ymax, vmax_ymax;			/* for index computation */
EXTERN long geom_vmax_zmax, geom_vmax_zmax_ymax, geom_vmax_ymax;	/* for geometry index computation */

EXTERN long DX, DY, DZ, DV;		/* index shifts */
EXTERN int dim,ONE,TWO,TRI;		/* used to control dimension */

/*
 *	Device Space defaults.
 */
#define SPACE_DEFAULT_X0 ONE
#define SPACE_DEFAULT_X1 (int)xmax-(1+ONE)
#define SPACE_DEFAULT_Y0 TWO
#define SPACE_DEFAULT_Y1 (int)ymax-(1+TWO)
#define SPACE_DEFAULT_Z0 TRI
#define SPACE_DEFAULT_Z1 (int)zmax-(1+TRI)

/**************************************************************************
 ************************** GEOMETRY **************************************
 **************************************************************************/

/*	Layers for geometry data.	*/
#define GEOM_STATUS  0
#define GEOM_FIBRE_1 1
#define GEOM_FIBRE_2 2
#define GEOM_FIBRE_3 3

/*	Point statuses.	*/
#define GEOM_VOID   0
#define GEOM_TISSUE 1
#define GEOM_TISSUE2 2

/**************************************************************************
 ************************** MPI ONLY **************************************
 **************************************************************************/
EXTERN int mpi_rank; /* in sequential version it will be zero */
#ifdef MPI
EXTERN int num_active_procs;
EXTERN MPI_Comm ALL_ACTIVE_PROCS;
EXTERN int I_AM_IDLE;
/*	Number of processors assigned to each axis.	*/
EXTERN int mpi_nx,mpi_ny,mpi_nz;
/*	Superindices for each parallel process.	*/
EXTERN int mpi_ix, mpi_iy, mpi_iz;
/*	Limits of a subdomain */
typedef struct {
	int local_xmin, local_xmax;
	int local_ymin, local_ymax;
	int local_zmin, local_zmax;
} Subdomain;
/*	Local minima and maxima for all subdomains.	*/
EXTERN Subdomain *subdomain_limits;

/* Subdomain details */
/*	Local bounds for each parallel process.	*/
EXTERN int local_xmin, local_xmax, local_ymin, local_ymax, local_zmin, local_zmax;
/*	Datatypes for sending hyperplanes.	*/
EXTERN MPI_Datatype XN_Type, XP_Type, YN_Type, YP_Type, ZN_Type, ZP_Type;
/*	Datatypes for receiving hyperplanes.	*/
EXTERN MPI_Datatype XN_Halo_Type, XP_Halo_Type, YN_Halo_Type, YP_Halo_Type, ZN_Halo_Type, ZP_Halo_Type;

/* Macros for calculating neighbouring processors
 * e.g. XN_NEIGHBOUR is neighbour in negative direction along x-axis,
 * 		YP_NEIGHBOUR is neighbour in positive direction along y-axis. */
int getProcessRank(int ix,int iy,int iz);
#define XN_NEIGHBOUR getProcessRank(mpi_ix-1,mpi_iy,mpi_iz) /*	neighbour to LEFT	*/
#define XP_NEIGHBOUR getProcessRank(mpi_ix+1,mpi_iy,mpi_iz) /*	neighbour to RIGHT	*/
#define YN_NEIGHBOUR getProcessRank(mpi_ix,mpi_iy-1,mpi_iz) /*	neighbour ABOVE	*/
#define YP_NEIGHBOUR getProcessRank(mpi_ix,mpi_iy+1,mpi_iz) /*	neighbour BELOW	*/
#define ZN_NEIGHBOUR getProcessRank(mpi_ix,mpi_iy,mpi_iz-1) /*	neighbour in FRONT	*/
#define ZP_NEIGHBOUR getProcessRank(mpi_ix,mpi_iy,mpi_iz+1) /*	neighbour BEHIND	*/

/**
 *	Get the rank of the process containing a 
 *	given internal or absolute boundary point.
 *
 */
int getRankContainingPoint(int x,int y,int z);

/* Macro to compute the index of a variable in New by its coordinates.
 * The MPI version has to be adjusted to match the continuous coordinate 
 * system to the local subdomain stored in New.
 *
 * Takes account of local bounds and whether or not boundary halos exist 
 * on each axis.
 */
#define ind(x,y,z,v) ( \
  ((x + ONE) - local_xmin)*vmax_zmax_ymax + \
  ((y + TWO) - local_ymin)*vmax_zmax + \
  ((z + TRI) - local_zmin)*vmax + \
  (v) \
)

/* Macro to compute the index of a variable in Geom by its coordinates.
 * The MPI version has to be adjusted to match the continuous coordinate system to
 * the local subdomain stored in New.
 *
 * Takes account of local bounds and whether or not boundary halos exist on each axis.
 */
#define geom_ind(x,y,z,v) ( \
  ((x + ONE) - local_xmin)*geom_vmax_zmax_ymax + \
  ((y + TWO) - local_ymin)*geom_vmax_zmax + \
  ((z + TRI) - local_zmin)*geom_vmax + \
  (v) \
)

#else
/*	Sequential version	*/
#define ind(x,y,z,v) ((x)*vmax_zmax_ymax+(y)*vmax_zmax+(z)*vmax+(v))
#define geom_ind(x,y,z,v) ((x)*geom_vmax_zmax_ymax+(y)*geom_vmax_zmax+(z)*geom_vmax+(v))
/* and these definition can help avoiding unnecessary #if(MPI) clauses: */
#define mpi_nx 1
#define mpi_ny 1
#define mpi_nz 1
#define mpi_ix 0
#define mpi_iy 0
#define mpi_iz 0
#endif

/* Tests the geometry status (TISSUE/VOID) of a point.
 * Takes (x,y,z) coordinates and evaluates to true or false.
 */
#define isTissue(x,y,z)	( \
  ( \
    x>=SPACE_DEFAULT_X0 && x<=SPACE_DEFAULT_X1 && \
    y>=SPACE_DEFAULT_Y0 && y<=SPACE_DEFAULT_Y1 && \
    z>=SPACE_DEFAULT_Z0 && z<=SPACE_DEFAULT_Z1 \
  ) && ( \
    !GEOMETRY_ON || \
    Geom[geom_ind(x,y,z,GEOM_STATUS)] == (real)GEOM_TISSUE || \
    Geom[geom_ind(x,y,z,GEOM_STATUS)] == (real)GEOM_TISSUE2 \
  ) \
)

/* indxy this is for screen only */
#define indxy(x,y) ((x)+xmax*(y))

/* Flag showing that some step procedure was already defined */
EXTERN int step_already_created;

int state(char *w);
void state_free(void);

#undef EXTERN

#endif /* _STATE_H_ */
